'******************************************************************************
'*  io.spin - handle input and output requests sent by pm80
'*
'*  (c) Juergen Buchmueller <pullmoll@t-online.de>
'*
'* $Id: io.spin,v 1.10 2010-03-16 17:11:54 pm Exp $
'******************************************************************************
CON

	io_cmd_out		=	$01		' I/O emulator command output
	io_cmd_in		=	$02		' I/O emulator command input
	io_cmd_break		=	$03		' I/O emulator command break

	constat			=	$10		' (16)sio port 1 status port
	condata			=	$11		' (17)sio port 1 data port
	punstat			=	$12		' (18)sio port 2 status port
	pundata			=	$13		' (19)sio port 2 data port

	myname_port		=	$70		' hopefully not used by anything else
	port2_Off		=	$71		' a little CP/M program sends an OUT to port &H71
	Forecolor		=	$72		' set foreground color
	Backcolor		=	$73		' set background color
	keystat			=	$74		' test just keyboard status (constat does this, too)
	keydata			=	$75		' get just the keyboard character

	port2_io		=	$78		' for input and output to port 2
	port2_test		=	$7D		' to test if a byte is ready
	port2_baud		=	$7C		' set baud rate multiple of 1200 eg 1=1200, 16=19200, 96=115200

	hdsk_reset		=	1		' command to reset controller
	hdsk_read		=	2		' read command
	hdsk_write		=	3		' write command
	hdsk_param		=	4		' get hard disk parameters
	hdsk_port		=	$fd		' control port for simulated hard disk

	simh_setbankselect	=	12		' SIMH command to set memory bank
	simh_reset		=	14		' SIMH reset command
	simh_hasbankedmemory	=	18		' SIMH command to check for banked memory
	simh_stoptimer		=	22		' SIMH stop timer interrupt
	simh_port		=	$fe		' control port for simulated hard disk

	SCR_W			=	40
	SCR_H			=	25

	serTX			=	31
	serRX			=	30
	baud			=	115200
	fdx_size		=	256

	_nul			=	0		' clear screen code
	_cr			=	13		' carriage return

VAR
	long	cog
	long	io_command
	long	io_port
	long	io_data
	long	addr

OBJ
#ifdef TV
	tv	: "Debug_1pinTV"
#endif
	fx	: "FullDuplexSerial_rr004"

PUB start(tvPin) : okay

	stop
	addr := CNT + clkfreq * 2	' wait 2 seconds
	waitcnt(addr)
	fx.start(serTX, serRX, 0, baud)
	addr :=  fx.params
	rx_head := addr
	rx_tail := addr + constant(1*4)
	tx_head := addr + constant(2*4)
	tx_tail := addr + constant(3*4)
	rx_buffer := addr + constant(9*4)
	tx_buffer := addr + constant(9*4) + constant(fdx_size)

#ifdef TV
        tv.start(tvPin)			'start the 1pinTV driver
	screen_ptr := tv.SCRPTR
#endif

	okay := cog := COGNEW(@entry, @io_command) + 1

PUB stop
'' Stop CPU simulator - frees a cog

	if cog
		cogstop(cog~ - 1)

PUB dec(value) | _i
'' Print a decimal number

  if value < 0
    -value
    fx.tx("-")

  _i := 1_000_000_000

  repeat 10
    if value => _i
      fx.tx(value / _i + "0")
      value //= _i
      result~~
    elseif result or _i == 1
      fx.tx("0")
    _i /= 10

PUB hex(value, digits)
'' Print a hexadecimal number

  value <<= (8 - digits) << 2
  repeat digits
    fx.tx(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PUB bin(value, digits)
'' Print a binary number

  value <<= 32 - digits
  repeat digits
    fx.tx((value <-= 1) & 1 + "0")

PUB command_addr
'' Return the address of the 3 longs block
	return @io_command

DAT
'---------------------------------------------------------------------------------------------------------

		org	0

'---------------------------------------------------------------------------------------------------------
entry
		mov	io_command_ptr, PAR
		mov	io_port_ptr, PAR
		add	io_port_ptr, #4
		mov	io_data_ptr, PAR
		add	io_data_ptr, #8
		jmp	#run

io_command_ptr	long	0-0
io_port_ptr	long	0-0
io_data_ptr	long	0-0

' FullDuplexSerial
rx_head		long	0-0
rx_tail		long	0-0
tx_head		long	0-0
tx_tail		long	0-0
rx_buffer	long	0-0
tx_buffer	long	0-0

wait		long	0
port		long	0
data		long	0
null		long	0

simh_command	long	0		' currently active command
simh_status	long	0		' status byte
simh_iobuf	long	0,0,0,0,0,0,0,0	' parameters of command
simh_iobuf_ptr	long	0		' pointer into simh_iobuf
simh_iobuf_cnt	long	0		' count of parameters expected

hdsk_command	long	0
hdsk_status	long	0
hdsk_iobuf	long	0,0,0,0,0,0,0,0	' parameters of command
hdsk_iobuf_ptr	long	0		' pointer into hdsk_iobuf
hdsk_iobuf_cnt	long	0		' count of parameters expected

'Hard Disk Parameter Blocks
'Each DPB is 17 bytes as defined for DPBs in CP/M followed by 2 bytes sector size as used by SIMH
hdsk_dpb
dpb_0                           ' 8MByte Default SIMH Altair HDSK params
spt_low		long	$20	' sectors per track (low byte)
spt_high	long	$00	' sectors per track (high byte)
bsh		long	$05	' data allocation Block SHift factor
blm		long	$1F	' data allocation block mask
exm		long	$01	' EXtent Mask
dsm_low		long	$F9	' maximum data block number (low_byte)
dsm_high	long	$07	' maximum data block number (high_byte)
drm_low		long	$FF	' total number of directory entries (low byte)
drm_high	long	$03	' total number of directory entries (high byte)
al0		long	$FF	' determine reserved directory blocks
al1		long	$00	' determine reserved directory blocks
cks_low		long	$00	' size of directory ChecK vector (low byte)
cks_high	long	$00	' size of directory ChecK vector (high byte)
off_low		long	$06	' number of reserved tracks (offset) (low byte)
off_high	long	$00	' number of reserved tracks (offset) (high byte)
psh		long	$00	' Physical record SHift factor, CP/M 3
phm		long	$00	' PHhysical record Mask, CP/M 3
ss_low		long	$80	' Sector Size (low byte)
ss_high		long	$00	' Sector Size (high byte)
dpb_end		long	$00

t1		long	0
t2		long	0
t3		long	0

#ifdef TV
screen_ptr	long	0
scr_x		long	0		' console cursor x
scr_y		long	1		' console cursor y
x20202020	long	$20202020
#endif

run
		wrlong	null, io_command_ptr
#ifdef TV
		call	#clear_screen
#endif
command
		rdlong	wait, io_command_ptr
		tjz	wait, #command
		cmp	wait, #io_cmd_out WZ
	if_z	jmp	#output
		cmp	wait, #io_cmd_in WZ
	if_z	jmp	#input
		cmp	wait, #io_cmd_break WZ
	if_z	jmp	#break
command_0
		wrlong	null, io_command_ptr
		jmp	#command

'*******************************************
'
' Read i/o port and data to write to
'
output
		rdlong	port, io_port_ptr
		rdlong	data, io_data_ptr
		wrlong	null, io_command_ptr
		and	port, #$ff
		cmp	port, #simh_port WZ	' SIMH port?
	if_z	jmp	#ouput_simh
		cmp	port, #hdsk_port WZ	' hard disk port?
	if_z	jmp	#ouput_hdsk
		cmp	port, #condata WZ	' console output?
	if_z	jmp	#output_conout
		jmp	#command

'*******************************************
'
' Write data to the console
'
output_conout
		and	data, #$ff
		call	#tx
#ifdef TV
		cmp	data, #_nul WZ
	if_z	jmp	#:cls
		cmp	data, #_cr WZ
	if_z	jmp	#:cr
		cmp	data, #32 WZ, WC
	if_b	jmp	#command
		jmp	#:emit
:cr
		call	#cursor_off
		mov	data, #32
		mov	scr_x, #0
		add	scr_y, #1
		cmp	scr_y, #SCR_H WZ, WC
	if_b	jmp	#:exit
		mov	scr_y, #(SCR_H - 1)
		call	#do_scroll
		jmp	#:exit
:emit
		call	#get_curpos
		wrbyte	data, t2
		add	scr_x, #1
		cmp	scr_x, #SCR_W WZ, WC	' < 40?
	if_ae	jmp	#:cr
:exit
		call	#cursor_on
		jmp	#command

:cls
		call	#clear_screen
#endif
		jmp	#command

#ifdef TV
do_scroll
		mov	t2, screen_ptr
		mov	t1, #SCR_W
		add	t1, t2
		mov	t3, #((SCR_H-1)*SCR_W/4)
:scroll
		rdlong	data, t1
		add	t1, #4
		wrlong	data, t2
		add	t2, #4
		djnz	t3, #:scroll
		mov	t3, #SCR_W/4
:erase_eol
		wrlong	x20202020, t2
		add	t2, #4
		djnz	t3, #:erase_eol
do_scroll_ret
		ret

clear_screen
		mov	t2, screen_ptr
		mov	t3, #(SCR_H*SCR_W/4)
:fill
		wrlong	x20202020, t2
		add	t2, #4
		djnz	t3, #:fill
		mov	scr_x, #0
		mov	scr_y, #0
clear_screen_ret
		ret

cursor_on
		call	#get_curpos
		mov	data, #2
		wrbyte	data, t2
cursor_on_ret
		ret

cursor_off
		call	#get_curpos
		mov	data, #32
		wrbyte	data, t2
cursor_off_ret
		ret

get_curpos
		mov	t1, scr_y
		mov	t2, #SCR_W
		call	#multiply		' y * SCR_W
		add	t2, scr_x		' + x
		add	t2, screen_ptr		' + screen base
get_curpos_ret
		ret
' multiply 16 bits t1 by t2, 32 bits result in t2; uses t3
multiply
		shl	t1, #16
		mov	t3, #16
		shr	t2, #1 WC
:loop	if_c	add	t2, t1 WC
		rcr	t2, #1 WC
		djnz	t3, #:loop
multiply_ret
		ret
#endif

'*******************************************
'
' Write data to the FullDuplexSerial tx_buffer
'
tx
		rdlong	t1, tx_head
		mov	t3, t1
		add	t1, #1
		rdlong	t2, tx_tail
		and	t1, #fdx_size-1
		cmp	t1, t2 WZ
	if_z	jmp	#tx			' wait until there's room in the buffer
		add	t3, tx_buffer
		wrbyte	data, t3
		wrlong	t1, tx_head		' increment the tx_head
tx_ret
		ret

'*******************************************
'
' Handle SIMH compatible command
'
ouput_simh
		tjnz	simh_iobuf_ptr, #output_simh_iobuf

		cmp	data, #simh_reset WZ	' reset what?
	if_z	jmp	#output_simh_reset
		cmp	data, #simh_hasbankedmemory WZ
	if_z	jmp	#output_simh_hasbankedmemory
		cmp	data, #simh_stoptimer WZ
	if_z	jmp	#output_simh_stoptimer
		cmp	data, #simh_setbankselect WZ
	if_z	jmp	#output_simh_setbankselect
		mov	simh_status, #$01	' set error status?!
		jmp	#command

output_simh_iobuf
		movd	:padr, simh_iobuf_ptr
		add	simh_iobuf_ptr, #1
:padr		mov	0-0, data		' store data in params
		sub	simh_iobuf_cnt, #1 WZ
	if_nz	jmp	#command
		mov	simh_iobuf_ptr, #0	' stop recording parameters
		' TODO: handle simh_command
		jmp	#command

output_simh_reset
		mov	simh_iobuf_ptr, #0	' stop recording or replying parameters
		mov	simh_iobuf_cnt, #0
		mov	simh_command, data
		mov	simh_status, #0
		jmp	#command

output_simh_setbankselect
		mov	simh_command, data
		mov	simh_iobuf_ptr, #simh_iobuf
		mov	simh_iobuf_cnt, #1	' expect one more output: the bank number
		jmp	#command

output_simh_hasbankedmemory
		mov	simh_command, data
		mov	simh_status, #0
		jmp	#command

output_simh_stoptimer
		mov	simh_command, data
		mov	simh_status, #0
		jmp	#command

'*******************************************
'
' Handle hard disk command
'
ouput_hdsk
		tjnz	hdsk_iobuf_ptr, #output_hdsk_iobuf

		cmp	data, #hdsk_reset WZ
	if_z	jmp	#output_hdsk_reset
		cmp	data, #hdsk_read WZ
	if_z	jmp	#output_hdsk_read
		cmp	data, #hdsk_write WZ
	if_z	jmp	#output_hdsk_write
		cmp	data, #hdsk_iobuf WZ
	if_z	jmp	#output_hdsk_params
		mov	hdsk_status, #$01	' set error status?!
		jmp	#command

output_hdsk_iobuf
		movd	:padr, hdsk_iobuf_ptr
		add	hdsk_iobuf_ptr, #1
:padr		mov	0-0, data		' store data in params
		sub	hdsk_iobuf_cnt, #1 WZ
	if_nz	jmp	#command
		mov	hdsk_iobuf_ptr, #0	' stop recording parameters
		' TODO: handle hdsk_command
		cmp	hdsk_command, #hdsk_read WZ
	if_z	jmp	#output_hdsk_read_dma
		cmp	hdsk_command, #hdsk_write WZ
	if_z	jmp	#output_hdsk_write_dma
		jmp	#command

output_hdsk_reset
		mov	hdsk_command, data
		mov	hdsk_iobuf_ptr, #0	' stop recording parameters
		mov	hdsk_iobuf_cnt, #0
		mov	hdsk_status, #0
		jmp	#command

output_hdsk_read
output_hdsk_write
		mov	hdsk_command, data
		mov	hdsk_iobuf_ptr, #hdsk_iobuf
		mov	hdsk_iobuf_cnt, #6	' drive, sector, track lo, track hi, DMA lo, DMA hi
		mov	hdsk_status, #1		' non-zero status while expecting parameters?
		jmp	#command

output_hdsk_params
		mov	hdsk_command, data
		mov	hdsk_iobuf_ptr, #hdsk_dpb
		mov	hdsk_iobuf_cnt, #19	' 19 bytes to read
		mov	hdsk_status, #0
		jmp	#command

output_hdsk_read_dma
		' TODO: read from the specified hard disk to the DMA area
		jmp	#command

output_hdsk_write_dma
		' TODO: write to the specified hard disk from the DMA area
		jmp	#command

'*******************************************
'
' Read i/o port to read from
'
input
		rdlong	port, io_port_ptr
		and	port, #$ff

		cmp	port, #simh_port WZ
	if_z	jmp	#input_simh
		cmp	port, #hdsk_port WZ
	if_z	jmp	#input_hdsk
		cmp	port, #constat WZ
	if_z	jmp	#input_constat
		cmp	port, #condata WZ
	if_z	jmp	#input_condata
		mov	data, #$00		' for now read $00 from all other ports
		wrlong	data, io_data_ptr
		jmp	#command_0

input_simh
		tjnz	simh_iobuf_ptr, #input_simh_iobuf
		wrlong	simh_status, io_data_ptr
		jmp	#command_0

input_simh_iobuf
		movs	:padr, simh_iobuf_ptr
		add	simh_iobuf_ptr, #1
:padr		mov	data, 0-0		' read params to data
		sub	hdsk_iobuf_cnt, #1 WZ
	if_nz	jmp	#command_0
		mov	simh_iobuf_ptr, #0	' stop replying parameters
		jmp	#command_0

input_hdsk
		tjnz	hdsk_iobuf_ptr, #input_hdsk_iobuf
		wrlong	hdsk_status, io_data_ptr
		jmp	#command_0

input_hdsk_iobuf
		movs	:padr, hdsk_iobuf_ptr
		add	hdsk_iobuf_ptr, #1
:padr		mov	data, 0-0		' read params to data
		sub	hdsk_iobuf_cnt, #1 WZ
	if_nz	jmp	#command_0
		mov	hdsk_iobuf_ptr, #0	' stop replying parameters
		jmp	#command_0

'*******************************************
'
' Read the console status
'
input_constat
		rdlong	t1, rx_head
		rdlong	t2, rx_tail
		cmp	t1, t2 WZ
	if_z	mov	data, #%00000010	' no data available
	if_z	wrlong	data, io_data_ptr
	if_z	jmp	#command_0
		mov	data, #%00000011	' data is available
		wrlong	data, io_data_ptr
		jmp	#command_0

input_condata
		rdlong	t1, rx_head
		rdlong	t2, rx_tail
		cmp	t1, t2 WZ
	if_z	mov	data, #0		' no data available
	if_z	wrlong	data, io_data_ptr
	if_z	jmp	#command_0
		add	t2, rx_buffer
		rdbyte	data, t2
		sub	t2, rx_buffer
		add	t2, #1
		and	t2, #fdx_size-1
		wrlong	t2, rx_tail
		wrlong	data, io_data_ptr	' store the input
		jmp	#command_0

'*******************************************
'
' Break - I don't know what this is good for
'
break
		jmp	#command_0

		fit	$1f0


